mod db_func;

use core::time;
use std::{fmt::{self, Display}, process::Command};
use regex::Regex;
use telegram_bot::{self, TelegramBot};
use redis;
use sea_orm::{self, sea_query::UnionType, IntoActiveModel, Set};
use tokio;
use tokio::time::{sleep, Duration};
use orm::sea_orm_active_enums::Status;

struct PingResults {
    sent: u16,
    received: u16,
    packet_loss: u8,
    total_time: u16,
    status: Status,
}

impl PingResults {
    fn new(sent: &str, received: &str, packet_loss: &str, time: &str) -> Self{
        let (sent_, received_, packet_loss_, time_ )= (
            u16::from_str_radix(sent, 10).unwrap_or(0),
            u16::from_str_radix(received, 10).unwrap_or(0),
            u8::from_str_radix(packet_loss, 10).unwrap_or(100),
            u16::from_str_radix(time, 10).unwrap_or(0),
        );

        let mut status: Status = Status::ConnError;

        if packet_loss_ <= 50 && 0 < time_ && time_ <= 5000 {
            status = Status::ConnGood
        }
        if 50 < packet_loss_ && packet_loss_ <= 75 || 5000 < time_ && time_ <= 10000 {
            status = Status::ConnProblems
        }

        return Self {
            sent: sent_,
            received: received_,
            packet_loss: packet_loss_,
            total_time: time_,
            status
        }
    }
    fn empty() -> Self {
        Self {
            sent: 0,
            received: 0,
            packet_loss: 0,
            total_time: 0,
            status: Status::ConnError
        }
    }
}

// enum UrlIPv4 {
//     Url(String),
//     IPv4(Ipv4Addr)
// }
struct Ping<'a> {
    dst_address: &'a str,
    count: u8,
    interval: f32,
    result: PingResults

}
impl<'a> Ping<'a>{
    fn new(dst_address: &'a str, count: u8, interval: f32) -> Self {

        let output = Command::new("ping")
            .arg(format!("-c {count}"))
            .arg(format!("-i {interval}"))
            .arg(format!("-w 5"))
            .arg(dst_address)
            .output();

        let re = Regex::new(
            r"---\n\D*(?P<tx>\d*)\D*(?P<rx>\d*)\D*(?P<pl>\d*)\D*(?P<time>\d*).*"
        ).unwrap();



        let result = match output {
            Ok(res) => {
                println!("stdout -> {:?}", String::from_utf8(res.stdout.clone()).unwrap());
                println!("stderr -> {:?}", res.stderr);
                let result = match String::from_utf8(res.stdout) {
                    Ok(res) => res,
                    Err(err) => panic!("{err}")
                };
                match re.captures(&result) {
                    Some(captures) => PingResults::new(
                        captures.name("tx").map_or("-1", |elem| elem.as_str()),
                        captures.name("rx").map_or("-1", |elem| elem.as_str()),
                        captures.name("pl").map_or("-1", |elem| elem.as_str()),
                        captures.name("time").map_or("-1", |elem| elem.as_str())
                    ),
                    None => PingResults::empty()
                }
            },
            Err(err) => panic!("{err}")
        };

        Self {
            dst_address, count, interval, result
        }
    }
    fn report(&self, location: &str) -> String{
        format!(
            "Изменение состояния интернет соединения в локации {}; \nРезультаты ping ->
            Потеря пакетов: {}%,
            Время отклика: {} мс
            Статус соединения: {}",
            location, self.result.packet_loss, self.result.total_time, self.result.status
        )
    }

}



impl Display for PingResults {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Sent: {}, Received: {}, Loss: {}, Time: {}, Status: {}",
            self.sent, self.received, self.packet_loss, self.total_time, self.status
        )
    }
}
#[tokio::main]
async fn main() {

    let db_url = std::env::var("DATABASE_URL").expect("DATABASE_URL is not set in .env file");
    let token = std::env::var("BOT_TOKEN").expect("BOT_TOKEN is not set in .env file");
    let redis_client: redis::Client = redis::Client::open("redis://127.0.0.1/").unwrap();

    let db_conn = sea_orm::Database::connect(db_url).await.unwrap();

    let bot = telegram_bot::TelegramBot::new(
        token, redis_client
    );

    const COUNT: u8 = 4;
    const INTERVAL: f32 = 0.2;

    let mut last_update_id: i64 = 0;

    loop {
        sleep(time::Duration::from_secs(60)).await;


        last_update_id = handle_updates(&bot, &db_conn, last_update_id).await;

        let locations = match db_func::get_locations(&db_conn).await {
            Ok(res) => res,
            Err(_) => continue,
        };

        for location in locations {
            let (location_data, user_tokens) = location;
            let ping = Ping::new(&location_data.endpoint, COUNT, INTERVAL);
            println!("{}", &ping.result.status);

            if location_data.status.unwrap_or(Status::ConnGood) != ping.result.status {

                match db_func::set_status_for_location(
                    &db_conn, ping.result.status, location_data.id
                ).await {
                    Ok(_) => (),
                    Err(err) => println!("{:}", err)
                };

                for user_token in user_tokens {
                    if user_token.chat_id.is_some() {
                        match bot.send_message(
                            user_token.chat_id.unwrap(),
                            &ping.report(&location_data.descr),
                            &telegram_bot::telegram_types::TelegramKeyboardMarkup { keyboard: vec![], resize_keyboard: false }
                        ).await {
                            Ok(_) => (),
                            Err(_) => continue
                        }
                    }
                }
            }
        }
    }

}

async fn handle_updates(
    bot: &TelegramBot, db: &sea_orm::DatabaseConnection, last_update_id: i64
) -> i64 {

    let mut __last_update_id: i64 = last_update_id;

    let regex_message = Regex::new(
        r"token:(?P<token>[0-9a-fA-F]{8}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{4}\b-[0-9a-fA-F]{12})"
    ).unwrap();

    let updates = match bot.get_updates(__last_update_id).await {
        Ok(res) => res,
        Err(_) => return last_update_id
    };

    for update in updates {
        if bot.check_if_handled(update.update_id).unwrap() {
            continue;
        }
        println!("{:?}", update.message.text);
        match regex_message.captures(&update.message.text) {
            Some(captures) => {
                match captures.name("token") {
                    Some(token) => {
                        match db_func::set_chat_id_by_token(
                            db, update.message.chat.id, token.as_str()
                        ).await {
                            Ok(_) => (),
                            Err(_) => continue
                        }
                    }
                    None => continue
                }
            },
            None => continue
        };
        bot.mark_as_handled(update.update_id).unwrap();
        __last_update_id = update.update_id
    }
    __last_update_id
}